"""$ fio load"""


from functools import partial
import itertools
import json
import logging

import click

import fiona
from fiona.fio import options
from fiona.transform import transform_geom


FIELD_TYPES_MAP_REV = dict([(v, k) for k, v in fiona.FIELD_TYPES_MAP.items()])


@click.command(short_help="Load GeoJSON to a dataset in another format.")
@click.argument('output', type=click.Path(), required=True)
@click.option('-f', '--format', '--driver', required=True,
              help="Output format driver name.")
@options.src_crs_opt
@click.option('--dst-crs', '--dst_crs',
              help="Destination CRS.  Defaults to --src-crs when not given.")
@click.option('--sequence / --no-sequence', default=False,
              help="Specify whether the input stream is a LF-delimited "
                   "sequence of GeoJSON features (the default) or a single "
                   "GeoJSON feature collection.")
@click.option('--layer', metavar="INDEX|NAME", callback=options.cb_layer,
              help="Load features into specified layer.  Layers use "
                   "zero-based numbering when accessed by index.")
@click.pass_context
def load(ctx, output, driver, src_crs, dst_crs, sequence, layer):
    """Load features from JSON to a file in another format.

    The input is a GeoJSON feature collection or optionally a sequence of
    GeoJSON feature objects."""
    verbosity = (ctx.obj and ctx.obj['verbosity']) or 2
    logger = logging.getLogger('fio')
    stdin = click.get_text_stream('stdin')

    dst_crs = dst_crs or src_crs

    if src_crs and dst_crs and src_crs != dst_crs:
        transformer = partial(transform_geom, src_crs, dst_crs,
                              antimeridian_cutting=True, precision=-1)
    else:
        transformer = lambda x: x

    first_line = next(stdin)

    # If input is RS-delimited JSON sequence.
    if first_line.startswith(u'\x1e'):
        def feature_gen():
            buffer = first_line.strip(u'\x1e')
            for line in stdin:
                if line.startswith(u'\x1e'):
                    if buffer:
                        feat = json.loads(buffer)
                        feat['geometry'] = transformer(feat['geometry'])
                        yield feat
                    buffer = line.strip(u'\x1e')
                else:
                    buffer += line
            else:
                feat = json.loads(buffer)
                feat['geometry'] = transformer(feat['geometry'])
                yield feat
    elif sequence:
        def feature_gen():
            yield json.loads(first_line)
            for line in stdin:
                feat = json.loads(line)
                feat['geometry'] = transformer(feat['geometry'])
                yield feat
    else:
        def feature_gen():
            text = "".join(itertools.chain([first_line], stdin))
            for feat in json.loads(text)['features']:
                feat['geometry'] = transformer(feat['geometry'])
                yield feat

    try:
        source = feature_gen()

        # Use schema of first feature as a template.
        # TODO: schema specified on command line?
        first = next(source)
        schema = {'geometry': first['geometry']['type']}
        schema['properties'] = dict([
            (k, FIELD_TYPES_MAP_REV.get(type(v)) or 'str')
            for k, v in first['properties'].items()])

        with fiona.drivers(CPL_DEBUG=verbosity > 2):
            with fiona.open(
                    output, 'w',
                    driver=driver,
                    crs=dst_crs,
                    schema=schema,
                    layer=layer) as dst:
                dst.write(first)
                dst.writerecords(source)

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



import fiona


def test_bounds_point():
    g = {'type': 'Point', 'coordinates': [10, 10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)


def test_bounds_line():
    g = {'type': 'LineString', 'coordinates': [[0, 0], [10, 10]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_polygon():
    g = {'type': 'Polygon', 'coordinates': [[[0, 0], [10, 10], [10, 0]]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_z():
    g = {'type': 'Point', 'coordinates': [10,10,10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)



import logging
import os.path
import shutil
import sys
import tempfile
import unittest

import fiona


logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)

FIXME_WINDOWS = sys.platform.startswith('win')

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Raises PermissionError Please look into why this test isn't working.")
def test_options(tmpdir=None):
    """Test that setting CPL_DEBUG=ON works"""
    if tmpdir is None:
        tempdir = tempfile.mkdtemp()
        logfile = os.path.join(tempdir, 'example.log')
    else:
        logfile = str(tmpdir.join('example.log'))
    logger = logging.getLogger('Fiona')
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(logfile)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    with fiona.drivers(CPL_DEBUG=True):
        c = fiona.open("tests/data/coutwildrnp.shp")
        c.close()
        log = open(logfile).read()
        assert "Option CPL_DEBUG" in log

    if tempdir and tmpdir is None:
        shutil.rmtree(tempdir)



import json
import re

import click
from click.testing import CliRunner

from fiona.fio import bounds

from .fixtures import (
    feature_collection, feature_collection_pp, feature_seq, feature_seq_pp_rs)


def test_fail():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, [], '5')
    assert result.exit_code == 1


def test_seq():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, [], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_seq_rs():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, [], feature_seq_pp_rs)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_precision():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--precision', 1], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d{1}\D', result.output)) == 8


def test_explode():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--explode'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_explode_pp():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--explode'], feature_collection_pp)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_with_id():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--with-id'], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('id') == result.output.count('bbox') == 2


def test_explode_with_id():
    runner = CliRunner()
    result = runner.invoke(
        bounds.bounds, ['--explode', '--with-id'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('id') == result.output.count('bbox') == 2


def test_with_obj():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--with-obj'], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('geometry') == result.output.count('bbox') == 2


def test_bounds_explode_with_obj():
    runner = CliRunner()
    result = runner.invoke(
        bounds.bounds, ['--explode', '--with-obj'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('geometry') == result.output.count('bbox') == 2


def test_explode_output_rs():
    runner = CliRunner()
    result = runner.invoke(bounds.bounds, ['--explode', '--rs'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count(u'\u001e') == 2
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8



import os
import tempfile

import fiona
import fiona.crs


def test_collection_crs_wkt():
    with fiona.open('tests/data/coutwildrnp.shp') as src:
        assert src.crs_wkt.startswith(
            'GEOGCS["GCS_WGS_1984",DATUM["WGS_1984",SPHEROID["WGS_84"')


def test_collection_no_crs_wkt():
    """crs members of a dataset with no crs can be accessed safely."""
    tmpdir = tempfile.gettempdir()
    filename = os.path.join(tmpdir, 'test.shp')
    with fiona.open('tests/data/coutwildrnp.shp') as src:
        profile = src.meta
    del profile['crs']
    del profile['crs_wkt']
    with fiona.open(filename, 'w', **profile) as dst:
        assert dst.crs_wkt == ""
        assert dst.crs == {}



"""Unittests for $ fio distrib"""


from click.testing import CliRunner

from fiona.fio import distrib

from .fixtures import feature_collection
from .fixtures import feature_collection_pp


def test_distrib():
    runner = CliRunner()
    result = runner.invoke(distrib.distrib, [], feature_collection_pp)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2


def test_distrib_no_rs():
    runner = CliRunner()
    result = runner.invoke(distrib.distrib, [], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2



import logging
import os
import shutil
import sys
import unittest

import fiona

logging.basicConfig(stream=sys.stderr, level=logging.INFO)

from .test_collection import ReadingTest


class VsiReadingTest(ReadingTest):
    
    # There's a bug in GDAL 1.9.2 http://trac.osgeo.org/gdal/ticket/5093
    # in which the VSI driver reports the wrong number of features.
    # I'm overriding ReadingTest's test_filter_1 with a function that
    # passes and creating a new method in this class that we can exclude
    # from the test runner at run time.

    def test_filter_vsi(self):
        results = list(self.c.filter(bbox=(-114.0, 35.0, -104, 45.0)))
        self.assertEqual(len(results), 67)
        f = results[0]
        self.assertEqual(f['id'], "0")
        self.assertEqual(f['properties']['STATE'], 'UT')


class ZipReadingTest(VsiReadingTest):
    
    def setUp(self):
        self.c = fiona.open("zip://tests/data/coutwildrnp.zip", "r")

    def tearDown(self):
        self.c.close()

    def test_open_repr(self):
        self.assertEqual(
            repr(self.c),
            ("<open Collection '/vsizip/tests/data/coutwildrnp.zip:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_closed_repr(self):
        self.c.close()
        self.assertEqual(
            repr(self.c),
            ("<closed Collection '/vsizip/tests/data/coutwildrnp.zip:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_path(self):
        self.assertEqual(self.c.path, '/vsizip/tests/data/coutwildrnp.zip')


class ZipArchiveReadingTest(VsiReadingTest):
    
    def setUp(self):
        self.c = fiona.open("/coutwildrnp.shp", "r", vfs="zip://tests/data/coutwildrnp.zip")
    
    def tearDown(self):
        self.c.close()

    def test_open_repr(self):
        self.assertEqual(
            repr(self.c),
            ("<open Collection '/vsizip/tests/data/coutwildrnp.zip/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_closed_repr(self):
        self.c.close()
        self.assertEqual(
            repr(self.c),
            ("<closed Collection '/vsizip/tests/data/coutwildrnp.zip/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_path(self):
        self.assertEqual(self.c.path, '/vsizip/tests/data/coutwildrnp.zip/coutwildrnp.shp')


class ZipArchiveReadingTestAbsPath(ZipArchiveReadingTest):

    def setUp(self):
        self.c = fiona.open(
                "/coutwildrnp.shp", "r",
                vfs="zip://" + os.path.abspath("tests/data/coutwildrnp.zip"))

    def test_open_repr(self):
        self.assert_(repr(self.c).startswith("<open Collection '/vsizip/"))

    def test_closed_repr(self):
        self.c.close()
        self.assert_(repr(self.c).startswith("<closed Collection '/vsizip/"))

    def test_path(self):
        self.assert_(self.c.path.startswith('/vsizip/'))


class TarArchiveReadingTest(VsiReadingTest):
    
    def setUp(self):
        self.c = fiona.open("/testing/coutwildrnp.shp", "r", vfs="tar://tests/data/coutwildrnp.tar")
    def tearDown(self):
        self.c.close()

    def test_open_repr(self):
        self.assertEqual(
            repr(self.c),
            ("<open Collection '/vsitar/tests/data/coutwildrnp.tar/testing/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_closed_repr(self):
        self.c.close()
        self.assertEqual(
            repr(self.c),
            ("<closed Collection '/vsitar/tests/data/coutwildrnp.tar/testing/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_path(self):
        self.assertEqual(self.c.path, '/vsitar/tests/data/coutwildrnp.tar/testing/coutwildrnp.shp')




import os.path


def read_file(name):
    return open(os.path.join(os.path.dirname(__file__), name)).read()

# GeoJSON feature collection on a single line
feature_collection = read_file('data/collection.txt')

# Same as above but with pretty-print styling applied
feature_collection_pp = read_file('data/collection-pp.txt')

# One feature per line
feature_seq = read_file('data/sequence.txt')

# Same as above but each feature has pretty-print styling
feature_seq_pp_rs = read_file('data/sequence-pp.txt')
